module jp.ku6.windows.game.animation;

import std.c.stdio;
import jp.ku6.windows.directx.d3d9types;
import jp.ku6.windows.directx.d3d9;
import jp.ku6.windows.game.graphics;

/++
 + 接触判定用矩形
 +/
struct CollisionRect
{
	/// 上端
	float top;

	/// 下端
	float bottom;

	/// 左端
	float left;

	/// 右端
	float right;

	/++
	 + 接触判定
	 +
	 + Params:
	 +  dx = 矩形の左上点X座標
	 +  dy = 矩形の左上点Y座標
	 +  sx = 判定対象矩形の左上点X座標
	 +  sy = 判定対象矩形の左上点Y座標
	 +  src = 判定対象矩形
	 +
	 + Returns:
	 +  接触している場合にtrueを返す
	 +/
	const
	bool isCollision(float dx, float dy, float sx, float sy, CollisionRect* src)
	nothrow
	{
		return src !is null && top + dy <= src.bottom + sy && bottom + dy >= src.top + sy && left + dx <= src.right + sx && right + dx >= src.left + sx;
	}
}

/++
 + 一定時間内の接触判定用データ
 +/
struct CollisionFrame
{
	// 使用終了時間
	ushort time;

	// 矩形の数
	ushort length;

	// 矩形データ
	CollisionRect* rects;

	/++
	 + 接触判定
	 +
	 + Params:
	 +  dx = X座標
	 +  dy = Y座標
	 +  sx = 判定対象のX座標
	 +  sy = 判定対象のY座標
	 +  src = 判定対象
	 +
	 + Returns:
	 +  接触している場合にtrueを返す
	 +/
	const
	bool isCollision(float dx, float dy, float sx, float sy, CollisionFrame* src)
	nothrow
	{
		for (ushort d; d < length; ++d)
		{
			for (ushort s; s < src.length; ++s)
			{
				if (rects[d].isCollision(dx, dy, sx, sy, &src.rects[s]))
					return true;
			}
		}
		return false;
	}
}

/++
 + 接触判定用データ
 +/
struct CollisionTimeline
{
	/// 接触判定用フレームの個数
	uint length;

	/// 接触判定用フレーム
	CollisionFrame* frames;
}


/++
 + アニメーション中で画像を組み合わせるためのパーツ
 +/
struct AnimationPart
{
	/// 表示タイプ
	uint type;

	/// パターン
	Pattern* pattern;

	/// 色
	uint color;

	/// 8byte alignのためのダミー
	uint dummy;

	/// 4頂点情報
	float[8] vertices;

	/++
	 + 描画
	 +
	 + Params:
	 +  x = 描画位置X座標
	 +  y = 描画位置Y座標
	 +/
	void draw(float x, float y)
	{
		if (pattern is null || pattern.surface is null || pattern.surface.texture is null)
			return;

		IDirect3DDevice9 device = Graphics.instance.direct3DDevice;

		switch (type)
		{
		case Sprite.DisplayType.NONE:
			return;
		case Sprite.DisplayType.NORMAL:
			device.SetRenderState(D3DRENDERSTATETYPE.D3DRS_DESTBLEND, D3DBLEND.D3DBLEND_INVSRCALPHA);
			break;
		case Sprite.DisplayType.ADD_BLEND:
			device.SetRenderState(D3DRENDERSTATETYPE.D3DRS_DESTBLEND, D3DBLEND.D3DBLEND_ONE);
			break;
		default:
			break;
		}

		Vertex[4] ver = void;

		// 左上
		ver[0].x = x + vertices[0];
		ver[0].y = y + vertices[1];
		ver[0].z = 0;
		ver[0].w = 1.0f;
		ver[0].color = color;
		ver[0].u = cast(float)(pattern.u) / cast(float)(pattern.surface.width);
		ver[0].v = cast(float)(pattern.v) / cast(float)(pattern.surface.height);

		// 右上
		ver[1].x = x + vertices[2];
		ver[1].y = y + vertices[3];
		ver[1].z = 0;
		ver[1].w = 1.0f;
		ver[1].color = ver[0].color;
		ver[1].u = cast(float)(pattern.u + pattern.width) / cast(float)(pattern.surface.width);
		ver[1].v = ver[0].v;

		// 右下
		ver[2].x = x + vertices[4];
		ver[2].y = y + vertices[5];
		ver[2].z = 0;
		ver[2].w = 1.0f;
		ver[2].color = ver[0].color;
		ver[2].u = ver[1].u;
		ver[2].v =cast(float)(pattern.v + pattern.height) / cast(float)(pattern.surface.height);

		// 左下
		ver[3].x = x + vertices[6];
		ver[3].y = y + vertices[7];
		ver[3].z = 0;
		ver[3].w = 1.0f;
		ver[3].color = ver[0].color;
		ver[3].u = ver[0].u;
		ver[3].v = ver[2].v;

		// テクスチャセット
		device.SetTexture(0, pattern.surface.texture);

		// レンダリング
		device.DrawPrimitiveUP(D3DPRIMITIVETYPE.D3DPT_TRIANGLEFAN, 2, &ver, Vertex.sizeof);
	}

private:

	static struct Vertex
	{
		float x,y,z,w;
		uint color;
		float u,v;
	}
}

/++
 + アニメーション中で画像パーツを組み合わせたもの
 +/
struct AnimationFrame
{
	/// 描画時間
	ushort time;

	/// パーツ数
	ushort length;

	/// パーツ
	AnimationPart* parts;

	/++
	 + 描画
	 +
	 + Params:
	 +  x = 描画位置X座標
	 +  y = 描画位置Y座標
	 +/
	void draw(float x, float y)
	{
		for (ushort i; i < length; ++i)
			parts[i].draw(x, y);
	}
}

/++
 + アニメーションデータ
 +/
struct AnimationTimeline
{
	// 描画フレーム数
	uint length;

	// 接触判定数
	uint collisionTimelineLength;

	// 描画フレーム
	AnimationFrame* animationFrames;

	// 接触判定
	CollisionTimeline* collisionTimelines;
}

/++
 + ファイル内の圧縮データをサーフェスに転送するクラス
 +/
class MemoryImage : SurfaceSource
{
	/++
	 + データをセット
	 +
	 + Params:
	 +  buffer = データポインタ
	 +  width = 画像の幅
	 +  height = 画像の高さ
	 +/
	void set(uint* buffer, uint width, uint height)
	{
		width_ = width;
		height_ = height;
		buffer_ = buffer;
	}

	uint width()
	{
		return width_;
	}

	uint height()
	{
		return height_;
	}

	string filePath()
	{
		return null;
	}

	bool open()
	{
		return true;
	}

	void close()
	{
	}

	void read(uint* buffer, uint width, uint height)
	{
		for (uint y; y < height_; ++y)
		{
			uint* d = &buffer[y * width];
			uint* s = &buffer_[y * width_];
			for (uint x; x < width_; ++x)
				d[x] = s[x];
		}
	}

private:

	// 画像の幅
	uint width_;

	// 画像の高さ
	uint height_;

	// データのポインタ
	uint* buffer_;
}

/++
 + アニメーションファイル
 +
 + ツールで作成したkagファイルをセットし、必要なデータを抜き出すためのインターフェース
 +/
class AnimationFile
{
	/++
	 + コンストラクタ
	 +/
	this()
	{
	}

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +/
	this(string filePath)
	{
		set(filePath);
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		release();
	}

	/++
	 + 内部で保持しているデータをクリアする
	 +/
	void release()
	{
		foreach (s; surfaceArray_)
			delete s;

		surfaceArray_ = null;
		patternArray_ = null;
		animationPartArray_ = null;
		animationFrameArray_ = null;
		animationTimelineArray_ = null;
		collisionRectArray_ = null;
		collisionFrameArray_ = null;
		collisionTimelineArray_ = null;
		surfaces_ = null;
		patterns_ = null;
		animationTimelines_ = null;
	}

	/++
	 + サーフェスを取得
	 +
	 + Params:
	 +  number = 識別番号
	 +
	 + Returns:
	 +  サーフェス
	 +  識別番号が間違っていた場合にはnullを返す
	 +/
	Surface getSurface(uint number)
	{
		Surface* surface = number in surfaces_;
		if (surface is null)
			return null;
		else
			return *surface;
	}

	/++
	 + パターンを取得
	 +
	 + Params:
	 +  number = 識別番号
	 +
	 + Returns:
	 +  パターン
	 +  識別番号が間違っていた場合にはnullを返す
	 +/
	Pattern* getPattern(uint number)
	{
		Pattern** pattern = number in patterns_;
		if (pattern is null)
			return null;
		else
			return *pattern;
	}

	/++
	 + アニメーションデータを取得
	 +
	 + Params:
	 +  number = 識別番号
	 +
	 + Returns:
	 +  アニメーションデータ
	 +  識別番号が間違っていた場合にはnullを返す
	 +/
	AnimationTimeline* getAnimationTimeline(uint number)
	{
		AnimationTimeline** timeline = number in animationTimelines_;
		if (timeline is null)
			return null;
		else
			return *timeline;
	}

	/++
	 + ファイルをセット
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +/
	void set(string filePath)
	{
		// 現在のデータは消去
		release();

		// ファイルをオープンする
		filePath ~= "\0";
		std.c.stdio.FILE* file = std.c.stdio.fopen(filePath.ptr, "rb\0".ptr);
		if (file is null)
			return;

		// ファイルヘッダを読み込み
		FileHeader header;
		std.c.stdio.fread(&header, FileHeader.sizeof, 1, file);


		// 圧縮画像読み込み用バッファ
		ubyte[] sourceBuffer = new ubyte[header.compressedImageMax];

		// 解凍画像書き込み用バッファ
		ubyte[] destinationBuffer = new ubyte[header.uncompressedImageMax];


		// 画像のヘッダ情報を取得
		SurfaceHeader[] surfaceHeaders = new SurfaceHeader[header.surfaceLength];
		std.c.stdio.fread(surfaceHeaders.ptr, SurfaceHeader.sizeof, surfaceHeaders.length, file);


		// 画像の識別番号を読み込み
		uint[] surfaceTags = new uint[header.surfaceLength];
		std.c.stdio.fread(surfaceTags.ptr, uint.sizeof, surfaceTags.length, file);


		// サーフェスを作成
		auto image = new MemoryImage;
		surfaceArray_ = new Surface[header.surfaceLength];
		for (uint i; i < header.surfaceLength; ++i)
		{
			// ピクセルデータをメモリ上にセット
			std.c.stdio.fread(sourceBuffer.ptr, ubyte.sizeof, surfaceHeaders[i].length, file);
			uncompress(sourceBuffer.ptr, surfaceHeaders[i].length, destinationBuffer.ptr);
			image.set(cast(uint*)destinationBuffer.ptr, surfaceHeaders[i].width, surfaceHeaders[i].height);

			// テクスチャを作成してセット
			surfaceArray_[i] = new Surface;
			surfaceArray_[i].load(image);
			surfaces_[surfaceTags[i]] = surfaceArray_[i];
		}

		// パターンの識別番号を読み込み
		uint[] patternTags = new uint[header.patternLength];
		std.c.stdio.fread(patternTags.ptr, uint.sizeof, patternTags.length, file);

		// パターンのロード
		int[] patternBytes = new int[header.patternLength * 5];
		std.c.stdio.fread(patternBytes.ptr, int.sizeof, patternBytes.length, file);

		patternArray_ = new Pattern[header.patternLength];			
		for (uint i; i < header.patternLength; ++i)
		{
			patternArray_[i].u = patternBytes[i * 5];
			patternArray_[i].v = patternBytes[i * 5 + 1];
			patternArray_[i].width = patternBytes[i * 5 + 2];
			patternArray_[i].height = patternBytes[i * 5 + 3];
			patternArray_[i].surface = surfaceArray_[patternBytes[i * 5 + 4]];

			patterns_[patternTags[i]] = &patternArray_[i];
		}

		// あたり判定はある？
		if ((header.flags & 0x00000002) != 0)
		{
			//	矩形範囲のバッファを作成
			collisionRectArray_ = new CollisionRect[header.collisionRectLength];
			std.c.stdio.fread(collisionRectArray_.ptr, CollisionRect.sizeof, collisionRectArray_.length, file);

			//	あたり判定のバッファを作成
			collisionFrameArray_ = new CollisionFrame[header.collisionFrameLength];
			std.c.stdio.fread(collisionFrameArray_.ptr, CollisionFrame.sizeof, collisionFrameArray_.length, file);

			//	あたり判定データのバッファを作成
			collisionTimelineArray_ = new CollisionTimeline[header.collisionTimelineLength];
			std.c.stdio.fread(collisionTimelineArray_.ptr, CollisionTimeline.sizeof, collisionTimelineArray_.length, file);
		}

		// アニメーションはある？
		if ((header.flags & 0x00000004) != 0)
		{
			// アニメーションパーツの個数を取得し、バッファを作成
			animationPartArray_ = new AnimationPart[header.animationPartLength];
			std.c.stdio.fread(animationPartArray_.ptr, AnimationPart.sizeof, animationPartArray_.length, file);

			// アニメーションフレームの個数を取得し、バッファを作成
			animationFrameArray_ = new AnimationFrame[header.animationFrameLength];
			std.c.stdio.fread(animationFrameArray_.ptr, AnimationFrame.sizeof, animationFrameArray_.length, file);

			// アニメーション識別番号領域を作成
			uint[] animationTagArray = new uint[header.animationTimelineLength];
			std.c.stdio.fread(animationTagArray.ptr, uint.sizeof, animationTagArray.length, file);

			// アニメーションデータの個数を取得し、バッファ作成
			animationTimelineArray_ = new AnimationTimeline[header.animationTimelineLength];
			std.c.stdio.fread(animationTimelineArray_.ptr, AnimationTimeline.sizeof, animationTimelineArray_.length, file);

			for (uint i; i < animationTimelineArray_.length; ++i)
				animationTimelines_[animationTagArray[i]] = &animationTimelineArray_[i];
		}

		// ファイルをクローズ
		std.c.stdio.fclose(file);

		foreach (ref f; collisionFrameArray_)
			f.rects = &collisionRectArray_[cast(uint)(f.rects)];

		foreach (ref t; collisionTimelineArray_)
			t.frames = &collisionFrameArray_[cast(uint)(t.frames)];

		foreach (ref p; animationPartArray_)
			p.pattern = &patternArray_[cast(uint)(p.pattern)];

		foreach (ref f; animationFrameArray_)
			f.parts = &animationPartArray_[cast(uint)(f.parts)];

		foreach (ref t; animationTimelineArray_)
		{
			t.animationFrames = &animationFrameArray_[cast(uint)(t.animationFrames)];
			if (cast(uint)t.collisionTimelines > 0)
				t.collisionTimelines = &collisionTimelineArray_[cast(uint)(t.collisionTimelines)];
		}
	}

private:

	/+
	 + ファイルヘッダ
	 +/
	struct FileHeader
	{
		// GA\0の文字
		char acTag[4];

		// 各種フラグ
		uint flags;

		// 内部で保持している圧縮済み画像の最大サイズ
		uint compressedImageMax;

		// 内部で保持している圧縮済み画像を展開した時の最大サイズ
		uint uncompressedImageMax;

		// テクスチャ数
		uint surfaceLength;

		// パターン数
		uint patternLength;

		// アニメーションパーツの数
		uint animationPartLength;

		// アニメーションフレームの数
		uint animationFrameLength;

		// アニメーションタイムラインの数
		uint animationTimelineLength;

		// 接触判定用矩形の数
		uint collisionRectLength;

		// 接触用データの数
		uint collisionFrameLength;

		// 接触用データのタイムラインの数
		uint collisionTimelineLength;
	}

	/+
	 + 
	 +/
	struct SurfaceHeader
	{
		// 幅
		ushort width;

		// 高さ
		ushort height;

		// データ長
		uint length;
	}

	// サーフェスサーフェスの配列
	Surface[] surfaceArray_;

	// パターンの配列
	Pattern[] patternArray_;

	// アニメーションパーツの配列
	AnimationPart[] animationPartArray_;

	// アニメーションフレームの配列
	AnimationFrame[] animationFrameArray_;

	// アニメーションデータの配列
	AnimationTimeline[] animationTimelineArray_;

	// 接触判定矩形の配列
	CollisionRect[] collisionRectArray_;

	// 接触判定フレームの配列
	CollisionFrame[] collisionFrameArray_;

	// 接触判定用データの配列
	CollisionTimeline[] collisionTimelineArray_;

	// サーフェスデータへのマップ
	Surface[uint] surfaces_;

	// パターンデータへのマップ
	Pattern*[uint] patterns_;

	// アニメーションデータへのマップ
	AnimationTimeline*[uint] animationTimelines_;

	/+
	 + ピクセルデータの解凍処理
	 +
	 + Params:
	 +  source = 圧縮データのポインタ
	 +  sourceLength = 圧縮データの長さ
	 +  destination = 解凍後のデータの格納領域
	 +/
	void uncompress(ubyte* source, uint sourceLength, ubyte* destination)
	{
		uint sourceCursor;
		uint destinationCursor;

		uint flags;
		while (sourceCursor < sourceLength)
		{
			// 圧縮フラグを読みきったら、圧縮フラグを1バイト取得
			flags >>= 1;
			if ((flags & 0x100) == 0 )
			{
				flags = source[sourceCursor++] | 0x0000FF00;
			}

			// 圧縮データか？
			if ((flags & 1) == 0)
			{
				// 一致位置と一致長を取得
				uint matchPosition = cast(uint)source[sourceCursor++];
				uint matchLength = cast(uint)source[sourceCursor++];
				matchPosition |= ((matchLength & 0xF0) << 4);
				matchLength = (matchLength & ~0xF0) + 3;

				// コピー元のデータのポインタを取得
				ubyte* matchedBytes = &destination[destinationCursor - matchPosition];

				// 一致長分コピー
				for (uint i; i < matchLength; ++i)
					destination[destinationCursor++] = matchedBytes[i];
			}
			else
			{
				// ふつーに出力
				destination[destinationCursor++] = source[sourceCursor++];
			}
		}
	}
}

/++
 + アニメーション
 +/
struct Animation
{
	/++
	 + 内部データを初期化する
	 +/
	void release()
	{
		counter_ = 0;
		currentFrame_ = 0;
		loopLength_ = 0;
		timeline_ = null;
		collisionCounts_ = null;
		collisionCurrentFrames_ = null;
	}

	/++
	 + 再生開始
	 +
	 + Params:
	 +  timeline = 再生データ
	 +  loopLength = ループ回数 負の値を指定すると無限ループ
	 +  restart = trueを指定すると再生中の場合でも最初から再生する
	 +/
	void play(AnimationTimeline* timeline, int loopLength, bool restart)
	{
		if (timeline is null)
		{
			counter_ = 0;
			return;
		}

		if (!restart && counter_ != 0 && timeline_ == timeline)
			return;

		// パラメータ設定
		counter_ = timeline.animationFrames[0].time;
		currentFrame_ = 0;
		loopLength_ = loopLength;
		timeline_ = timeline;

		uint collisionLength = timeline.collisionTimelines is null ? 0 : timeline.collisionTimelines.length;
		collisionCounts_.length = collisionLength;
		collisionCurrentFrames_.length = collisionLength;
		for (uint i; i < collisionLength; ++i)
		{
			collisionCounts_[i] = timeline.collisionTimelines[i].frames[0].time;
			collisionCurrentFrames_[i] = 0;
		}
	}

	/++
	 + 停止
	 +/
	void stop()
	{
		counter_ = 0;
	}

	/++
	 + 衝突判定データを取得
	 +
	 + Params:
	 +  index = インデックス
	 +/
	CollisionFrame*	getCurrentCollisionFrame(uint index)
	{
		if (timeline_ is null || timeline_.collisionTimelines is null || timeline_.collisionTimelines.length <= index)
			return null;
		else
			return &(timeline_.collisionTimelines[index].frames[collisionCurrentFrames_[index]]);
	}

	/++
	 + 毎フレームの処理
	 +
	 + アニメーションをさせるために毎フレーム必ず呼び出すこと
	 + これを呼び出さないとアニメーションは停止したままとなる
	 +/
	void move()
	{
		if (timeline_ is null)
			return;

		if (counter_ == 0)
			return;

		// アニメーションフレームの表示時間と比較
		if (--counter_ == 0)
		{
			// 新しいフレームに切り替える
			if (++currentFrame_ >= timeline_.length)
			{
				if (loopLength_ > 0)
					--loopLength_;

				if (loopLength_ != 0)
					currentFrame_ = 0;
				else
					--currentFrame_; // 最終フレームで固定
			}

			// カウンタを再設定
			counter_ = timeline_.animationFrames[currentFrame_].time;
		}

		uint collisionLength = timeline_.collisionTimelines is null ? 0 : timeline_.collisionTimelines.length;
		for (uint i; i < collisionLength; ++i)
		{
			if (collisionCounts_[i] != 0 && --collisionCounts_[i] == 0)
			{
				// 新しいフレームに切り替える
				if (++collisionCurrentFrames_[i] >= timeline_.collisionTimelines[i].frames.length)
				{
					if (loopLength_ != 0)
						collisionCurrentFrames_[i] = 0;
					else
						--collisionCurrentFrames_[i]; // 最終フレームで固定
				}

				// カウンタを再設定
				collisionCounts_[i] = timeline_.collisionTimelines[i].frames[collisionCurrentFrames_[i]].time;
			}
		}
	}

	/++
	 + 描画処理
	 +
	 + Params:
	 +  x = 描画位置X座標
	 +  y = 描画位置Y座標
	 +/
	void draw(float x, float y)
	{
		if (timeline_ !is null)
			timeline_.animationFrames[currentFrame_].draw(x, y);
	}

private:

	// 再生時間のカウンタ
	short counter_;

	// 表示中のフレーム
	int currentFrame_;

	// ループ回数 (負の数を指定すると無限ループ)
	int loopLength_;

	// 再生するアニメーション
	AnimationTimeline* timeline_;

	// 当り判定の再生時間カウンタ
	int[] collisionCounts_;

	// 当り判定の使用中フレーム
	int[] collisionCurrentFrames_;
}

